﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

using Hjg.Pngcs;
using Hjg.Pngcs.Chunks;

namespace Automap
{
	public class Snapshotter
	{
		private const string customTimeFormat = @"yyyy.MM.dd.HH.mm.ssZzz";
		public readonly int chunkSize;
		public readonly int worldSeed;
		public readonly string fileName;
		public readonly string chunkPath;
		public readonly int NorthEdge;
		public readonly int WestEdge;
		public readonly int Width;
		public readonly int Height;

		public ColumnsMetadata colMetadata;
		public bool Imaging { get; private set; }
		public bool Finished { get; private set; }
		public Stopwatch Timer { get; private set;}

		/// <summary>
		/// Prepares class to take "snapshot" (single use!)
		/// </summary>
		/// <param name="path">path to the map dir</param>
		/// <param name="cols">Input metadata</param>
		/// <param name="chunkSizeP">size in blocks of std. chunk</param>
		/// <param name="worldSeedP">World Seed #</param>
		public Snapshotter(string path, ColumnsMetadata cols, int chunkSizeP, int worldSeedP)
		{				
			colMetadata = cols;
			chunkSize = chunkSizeP;
			worldSeed = worldSeedP;
			NorthEdge = cols.North_mostChunk;
			WestEdge = cols.West_mostChunk;
			Width = (cols.East_mostChunk - WestEdge + 1);
			Height = (cols.South_mostChunk - NorthEdge + 1);
			fileName = Path.Combine(path, $"snapshot_{worldSeed}_{DateTime.Now.ToString(customTimeFormat)}.png");
			chunkPath = Path.Combine(path, AutomapSystem._chunkPath);
		}

		/// <summary>
		/// takes a snapshot. this should be called from an extra thread.
		/// </summary>
		public async void Take()
		{
			if (Imaging || Finished) return;//Can't take annother photo...this one is started or complete
			Imaging = true;
			Timer = new Stopwatch();
			Timer.Start();

			Debug.WriteLine("snapshot started");

			ImageInfo info = new ImageInfo(Width * chunkSize, Height * chunkSize, 8, false);
			PngWriter snapWriter = FileHelper.CreatePngWriter(fileName, info, true);
			PngMetadata meta = snapWriter.GetMetadata( );
			meta.SetTimeNow( );
			var transparencyChunk = meta.CreateTRNSChunk( );
			transparencyChunk.SetRGB(0, 0, 0);//CHECK: This is pure black.
			meta.SetText("Northmost", NorthEdge.ToString("D"));
			meta.SetText("Eastmost", WestEdge.ToString("D"));
			meta.SetText("Width", Width.ToString("D"));
			meta.SetText("Height", Height.ToString("D"));


		
			/*
		 	Red:   2 bytes, range 0 .. (2^bitdepth)-1
			Green: 2 bytes, range 0 .. (2^bitdepth)-1
			Blue:  2 bytes, range 0 .. (2^bitdepth)-1
			*/


			snapWriter.CompLevel = 5;
			snapWriter.CompressionStrategy = Hjg.Pngcs.Zlib.EDeflateCompressStrategy.Filtered;



			var orderedList =
				from ch in colMetadata
				group ch by ch.Location.Y into g
				orderby g.Key
				select g;
			// that sorts things in ascending order so we can only create (chunkSize) lines at once
			int row = 0, lastPosY = 0, posY = 0;
			foreach (var chunkGroup in orderedList)
			{
				// oh god here we go...
				posY = chunkGroup.Key - NorthEdge;
				var delta = posY - lastPosY;
				lastPosY = posY;

				// if there is more than 1 blank step then the gap needs to be filled.
				for (int i = 1; i < delta; i++)
				{
					byte[] blankLine = new byte[info.BytesPerRow];
					for (int j = 0; j < chunkSize; j++)
						snapWriter.WriteRowByte(blankLine, row++);
				}

				var inGroup = (await ReadAllInGroup(chunkGroup)).ToArray();
				
				for (int r = 0; r < chunkSize; r++)
				{
					var line = new byte[info.BytesPerRow];
					for (int chunk = 0; chunk < inGroup.Length; chunk++)
					{
						var posX = inGroup[chunk].Key * chunkSize * info.BytesPixel;
						inGroup[chunk].Value?[r].CopyTo(line, posX);
					}
					snapWriter.WriteRowByte(line, row++);
				}
			}
			snapWriter.ShouldCloseStream = true;
			try
			{
				snapWriter.End();
				Timer.Stop( );
				Imaging = false;
				Finished = true;
			}
			catch (Exception)
			{
				Debug.WriteLine("Snapshot exception!");
			}
			Debug.WriteLine($"snapshot finished in {Timer.ElapsedMilliseconds}");
		}

		private async Task<Dictionary<int, byte[][]>> ReadAllInGroup(IGrouping<int, ColumnMeta> group)
		{
			var taskGroup = new Dictionary<int, byte[][]>(group.Count());
			foreach (var shardMeta in group)
			{
				var shardPath = Path.Combine(chunkPath, $"{shardMeta.Location.X}_{shardMeta.Location.Y}.png");
				taskGroup.Add(shardMeta.Location.X - WestEdge, await ReadNoThrow(shardPath));
			}
			return taskGroup;
		}

		private async Task<byte[][]> ReadNoThrow(string readPath)
		{
			try
			{
				return await Task.Run(() => FileHelper.CreatePngReader(readPath).ReadRowsByte().ScanlinesB);
			}
			catch (Exception e)
			{
				Debug.WriteLine(e.Message);
				return null;
			} // do nothing on error
		}
	}
}
